// Copyright (C) 2021  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package blib
// set of [lo,hi) Key ranges.

//go:generate ./gen-rangemap _RangedMap_void void zrangemap_void.go

// RangedKeySet is set of Keys with adjacent keys coalesced into Ranges.
//
// Zero value represents empty set.
type RangedKeySet struct {
	m  _RangedMap_void // XXX naming
}


// void is used as value type for RangedMap to be used as set.
type void struct{}
func (_ void) String() string { return "" }


// Union returns RangedKeySet(A.keys | B.keys).
func (A *RangedKeySet) Union(B *RangedKeySet) *RangedKeySet {
	U := A.Clone()
	U.UnionInplace(B)
	return U
}

// Difference returns RangedKeySet(A.keys \ B.keys).
func (A *RangedKeySet) Difference(B *RangedKeySet) *RangedKeySet {
	D := A.Clone()
	D.DifferenceInplace(B)
	return D
}

// XXX Intersection

func (A *RangedKeySet) UnionInplace(B *RangedKeySet) {
	A.verify()
	B.verify()
	defer A.verify()

	// XXX dumb
	for _, r := range B.rangev {
		A.AddRange(r)
	}
}

func (A *RangedKeySet) DifferenceInplace(B *RangedKeySet) {
	A.verify()
	B.verify()
	defer A.verify()

	// XXX dumb
	for _, r := range B.rangev {
		if len(A.rangev) == 0 {
			break
		}
		A.DelRange(r)
	}
}
